@@ -13,6 +13,7 @@
 
 """Implementation of SQLAlchemy backend."""
 import datetime
+import functools
 import itertools
 import random
 
@@ -24,9 +25,9 @@ from oslo_db.sqlalchemy import enginefacade
 from oslo_db.sqlalchemy import utils
 from oslo_log import log as logging
 from oslo_utils import encodeutils
+from oslo_utils import excutils
 from oslo_utils import timeutils
 import osprofiler.sqlalchemy
-import six
 import sqlalchemy
 from sqlalchemy import and_
 from sqlalchemy import func
@@ -48,6 +49,12 @@ CONF = cfg.CONF
 CONF.import_opt('hidden_stack_tags', 'heat.common.config')
 CONF.import_opt('max_events_per_stack', 'heat.common.config')
 CONF.import_group('profiler', 'heat.common.config')
+CONF.import_opt('db_max_retries', 'oslo_db.options', group='database')
+CONF.import_opt('db_retry_interval', 'oslo_db.options', group='database')
+CONF.import_opt(
+    'db_inc_retry_interval', 'oslo_db.options', group='database')
+CONF.import_opt(
+    'db_max_retry_interval', 'oslo_db.options', group='database')
 
 options.set_defaults(CONF)
 
@@ -58,7 +65,9 @@ LOG = logging.getLogger(__name__)
 
 
 # TODO(sbaker): fix tests so that sqlite_fk=True can be passed to configure
-db_context.configure()
+# FIXME(stephenfin): we need to remove reliance on autocommit semantics ASAP
+# since it's not compatible with SQLAlchemy 2.0
+db_context.configure(__autocommit=True)
 
 
 def get_facade():
@@ -88,9 +97,33 @@ def get_session():
     return get_facade().get_session()
 
 
+def retry_on_db_error(func):
+    @functools.wraps(func)
+    def try_func(context, *args, **kwargs):
+        if (context.session.transaction is None or
+                not context.session.autocommit):
+            wrapped = oslo_db_api.wrap_db_retry(
+                max_retries=CONF.database.db_max_retries,
+                retry_on_deadlock=True,
+                retry_on_disconnect=True,
+                retry_interval=CONF.database.db_retry_interval,
+                inc_retry_interval=CONF.database.db_inc_retry_interval,
+                max_retry_interval=CONF.database.db_max_retry_interval)(func)
+            return wrapped(context, *args, **kwargs)
+        else:
+            try:
+                return func(context, *args, **kwargs)
+            except (db_exception.DBDeadlock, db_exception.DBConnectionError):
+                with excutils.save_and_reraise_exception():
+                    LOG.debug('Not retrying on DBDeadlock and '
+                              'DBConnectionError because '
+                              'transaction not closed')
+    return try_func
+
+
 def update_and_save(context, obj, values):
     with context.session.begin(subtransactions=True):
-        for k, v in six.iteritems(values):
+        for k, v in values.items():
             setattr(obj, k, v)
 
 
@@ -142,7 +175,11 @@ def raw_template_update(context, template_id, values):
 
 
 def raw_template_delete(context, template_id):
-    raw_template = raw_template_get(context, template_id)
+    try:
+        raw_template = raw_template_get(context, template_id)
+    except exception.NotFound:
+        # Ignore not found
+        return
     raw_tmpl_files_id = raw_template.files_id
     session = context.session
     with session.begin(subtransactions=True):
@@ -153,7 +190,12 @@ def raw_template_delete(context, template_id):
         # delete that too
         if session.query(models.RawTemplate).filter_by(
                 files_id=raw_tmpl_files_id).first() is None:
-            raw_tmpl_files = raw_template_files_get(context, raw_tmpl_files_id)
+            try:
+                raw_tmpl_files = raw_template_files_get(
+                    context, raw_tmpl_files_id)
+            except exception.NotFound:
+                # Ignore not found
+                return
             session.delete(raw_tmpl_files)
 
 
@@ -234,14 +276,13 @@ def resource_get_all(context):
     return results
 
 
-@oslo_db_api.wrap_db_retry(max_retries=3, retry_on_deadlock=True,
-                           retry_interval=0.5, inc_retry_interval=True)
+@retry_on_db_error
 def resource_purge_deleted(context, stack_id):
     filters = {'stack_id': stack_id, 'action': 'DELETE', 'status': 'COMPLETE'}
     query = context.session.query(models.Resource)
     result = query.filter_by(**filters)
     attr_ids = [r.attr_data_id for r in result if r.attr_data_id is not None]
-    with context.session.begin(subtransactions=True):
+    with context.session.begin():
         result.delete()
         if attr_ids:
             context.session.query(models.ResourcePropertiesData).filter(
@@ -256,10 +297,15 @@ def _add_atomic_key_to_values(values, atomic_key):
         values['atomic_key'] = atomic_key + 1
 
 
-@oslo_db_api.wrap_db_retry(max_retries=3, retry_on_deadlock=True,
-                           retry_interval=0.5, inc_retry_interval=True)
+@retry_on_db_error
 def resource_update(context, resource_id, values, atomic_key,
                     expected_engine_id=None):
+    return _try_resource_update(context, resource_id, values, atomic_key,
+                                expected_engine_id)
+
+
+def _try_resource_update(context, resource_id, values, atomic_key,
+                         expected_engine_id=None):
     session = context.session
     with session.begin(subtransactions=True):
         _add_atomic_key_to_values(values, atomic_key)
@@ -289,7 +335,7 @@ def resource_delete(context, resource_id):
 
 def resource_attr_id_set(context, resource_id, atomic_key, attr_id):
     session = context.session
-    with session.begin(subtransactions=True):
+    with session.begin():
         values = {'attr_data_id': attr_id}
         _add_atomic_key_to_values(values, atomic_key)
         rows_updated = session.query(models.Resource).filter(and_(
@@ -304,7 +350,7 @@ def resource_attr_id_set(context, resource_id, atomic_key, attr_id):
         else:
             # Someone else set the attr_id first and/or we have a stale
             # view of the resource based on atomic_key, so delete the
-            # resource_properties_data (attr) db row.
+            # resource_properties_data (attr) DB row.
             LOG.debug('Not updating res_id %(rid)s with attr_id %(aid)s',
                       {'rid': resource_id, 'aid': attr_id})
             session.query(
@@ -315,7 +361,7 @@ def resource_attr_id_set(context, resource_id, atomic_key, attr_id):
 
 def resource_attr_data_delete(context, resource_id, attr_id):
     session = context.session
-    with session.begin(subtransactions=True):
+    with session.begin():
         resource = session.query(models.Resource).get(resource_id)
         attr_prop_data = session.query(
             models.ResourcePropertiesData).get(attr_id)
@@ -341,9 +387,14 @@ def resource_data_get_all(context, resource_id, data=None):
 
     for res in data:
         if res.redact:
-            ret[res.key] = crypt.decrypt(res.decrypt_method, res.value)
-        else:
-            ret[res.key] = res.value
+            try:
+                ret[res.key] = crypt.decrypt(res.decrypt_method, res.value)
+                continue
+            except exception.InvalidEncryptionKey:
+                LOG.exception('Failed to decrypt resource data %(rkey)s '
+                              'for %(rid)s, ignoring.',
+                              {'rkey': res.key, 'rid': resource_id})
+        ret[res.key] = res.value
     return ret
 
 
@@ -437,7 +488,7 @@ def resource_exchange_stacks(context, resource_id1, resource_id2):
 def resource_data_delete(context, resource_id, key):
     result = resource_data_get_by_key(context, resource_id, key)
     session = context.session
-    with session.begin(subtransactions=True):
+    with session.begin():
         session.delete(result)
 
 
@@ -448,20 +499,20 @@ def resource_create(context, values):
     return resource_ref
 
 
+@retry_on_db_error
 def resource_create_replacement(context,
-                                existing_res_id, existing_res_values,
+                                existing_res_id,
                                 new_res_values,
                                 atomic_key, expected_engine_id=None):
     session = context.session
     try:
-        with session.begin(subtransactions=True):
+        with session.begin():
             new_res = resource_create(context, new_res_values)
             update_data = {'replaced_by': new_res.id}
-            update_data.update(existing_res_values)
-            if not resource_update(context,
-                                   existing_res_id, update_data,
-                                   atomic_key,
-                                   expected_engine_id=expected_engine_id):
+            if not _try_resource_update(context,
+                                        existing_res_id, update_data,
+                                        atomic_key,
+                                        expected_engine_id=expected_engine_id):
                 data = {}
                 if 'name' in new_res_values:
                     data['resource_name'] = new_res_values['name']
@@ -571,7 +622,7 @@ def stack_get_by_name(context, stack_name):
              models.Stack.tenant == context.tenant_id,
              models.Stack.stack_user_project_id == context.tenant_id)
              ).filter_by(name=stack_name)
-    return query.first()
+    return query.order_by(models.Stack.created_at).first()
 
 
 def stack_get(context, stack_id, show_deleted=False, eager_load=True):
@@ -621,13 +672,13 @@ def stack_get_all_by_root_owner_id(context, owner_id):
 
 
 def _get_sort_keys(sort_keys, mapping):
-    """Returns an array containing only whitelisted keys
+    """Returns an array containing only allowed keys
 
     :param sort_keys: an array of strings
     :param mapping: a mapping from keys to DB column names
     :returns: filtered list of sort keys
     """
-    if isinstance(sort_keys, six.string_types):
+    if isinstance(sort_keys, str):
         sort_keys = [sort_keys]
     return [mapping[key] for key in sort_keys or [] if key in mapping]
 
@@ -656,7 +707,7 @@ def _paginate_query(context, query, model, limit=None, sort_keys=None,
     return query
 
 
-def _query_stack_get_all(context,  show_deleted=False,
+def _query_stack_get_all(context, show_deleted=False,
                          show_nested=False, show_hidden=False, tags=None,
                          tags_any=None, not_tags=None, not_tags_any=None):
     if show_nested:
@@ -733,11 +784,11 @@ def _filter_and_page_query(context, query, limit=None, sort_keys=None,
                     rpc_api.STACK_STATUS: models.Stack.status.key,
                     rpc_api.STACK_CREATION_TIME: models.Stack.created_at.key,
                     rpc_api.STACK_UPDATED_TIME: models.Stack.updated_at.key}
-    whitelisted_sort_keys = _get_sort_keys(sort_keys, sort_key_map)
+    valid_sort_keys = _get_sort_keys(sort_keys, sort_key_map)
 
     query = db_filters.exact_filter(query, models.Stack, filters)
     return _paginate_query(context, query, models.Stack, limit,
-                           whitelisted_sort_keys, marker, sort_dir)
+                           valid_sort_keys, marker, sort_dir)
 
 
 def stack_count_all(context, filters=None,
@@ -757,12 +808,21 @@ def stack_count_all(context, filters=None,
 def stack_create(context, values):
     stack_ref = models.Stack()
     stack_ref.update(values)
+    stack_name = stack_ref.name
     stack_ref.save(context.session)
+
+    # Even though we just created a stack with this name, we may not find
+    # it again because some unit tests create stacks with deleted_at set. Also
+    # some backup stacks may not be found, for reasons that are unclear.
+    earliest = stack_get_by_name(context, stack_name)
+    if earliest is not None and earliest.id != stack_ref.id:
+        context.session.query(models.Stack).filter_by(id=stack_ref.id).delete()
+        raise exception.StackExists(stack_name=stack_name)
+
     return stack_ref
 
 
-@oslo_db_api.wrap_db_retry(max_retries=3, retry_on_deadlock=True,
-                           retry_interval=0.5, inc_retry_interval=True)
+@retry_on_db_error
 def stack_update(context, stack_id, values, exp_trvsl=None):
     session = context.session
     with session.begin(subtransactions=True):
@@ -820,7 +880,9 @@ def _is_duplicate_error(exc):
 
 
 @oslo_db_api.wrap_db_retry(max_retries=3, retry_on_deadlock=True,
-                           retry_interval=0.5, inc_retry_interval=True,
+                           retry_on_disconnect=True,
+                           retry_interval=0.5,
+                           inc_retry_interval=True,
                            exception_checker=_is_duplicate_error)
 def stack_lock_create(context, stack_id, engine_id):
     with db_context.writer.independent.using(context) as session:
@@ -906,7 +968,7 @@ def user_creds_create(context):
     else:
         user_creds_ref.update(values)
         method, password = crypt.encrypt(values['password'])
-        if len(six.text_type(password)) > 255:
+        if len(str(password)) > 255:
             raise exception.Error(_("Length of OS_PASSWORD after encryption"
                                     " exceeds Heat limit (255 chars)"))
         user_creds_ref.password = password
@@ -926,7 +988,7 @@ def user_creds_get(context, user_creds_id):
     db_result = context.session.query(models.UserCreds).get(user_creds_id)
     if db_result is None:
         return None
-    # Return a dict copy of db results, do not decrypt details into db_result
+    # Return a dict copy of DB results, do not decrypt details into db_result
     # or it can be committed back to the DB in decrypted form
     result = dict(db_result)
     del result['decrypt_method']
@@ -1011,12 +1073,12 @@ def _events_filter_and_page_query(context, query,
 
     sort_key_map = {rpc_api.EVENT_TIMESTAMP: models.Event.created_at.key,
                     rpc_api.EVENT_RES_TYPE: models.Event.resource_type.key}
-    whitelisted_sort_keys = _get_sort_keys(sort_keys, sort_key_map)
+    valid_sort_keys = _get_sort_keys(sort_keys, sort_key_map)
 
     query = db_filters.exact_filter(query, models.Event, filters)
 
     return _events_paginate_query(context, query, models.Event, limit,
-                                  whitelisted_sort_keys, marker, sort_dir)
+                                  valid_sort_keys, marker, sort_dir)
 
 
 def event_count_all_by_stack(context, stack_id):
@@ -1084,7 +1146,7 @@ def _delete_event_rows(context, stack_id, limit):
         (ids, rsrc_prop_ids) = zip(*id_pairs)
         max_id = ids[-1]
         # delete the events
-        retval = session.query(models.Event.id).filter(
+        retval = session.query(models.Event).filter(
             models.Event.id <= max_id).filter(
                 models.Event.stack_id == stack_id).delete()
 
@@ -1118,6 +1180,7 @@ def _delete_event_rows(context, stack_id, limit):
     return retval
 
 
+@retry_on_db_error
 def event_create(context, values):
     if 'stack_id' in values and cfg.CONF.max_events_per_stack:
         # only count events and purge on average
@@ -1131,7 +1194,7 @@ def event_create(context, values):
                 _delete_event_rows(context, values['stack_id'],
                                    cfg.CONF.event_purge_batch_size)
             except db_exception.DBError as exc:
-                LOG.error('Failed to purge events: %s', six.text_type(exc))
+                LOG.error('Failed to purge events: %s', str(exc))
     event_ref = models.Event()
     event_ref.update(values)
     event_ref.save(context.session)
@@ -1225,7 +1288,7 @@ def software_deployment_update(context, deployment_id, values):
 def software_deployment_delete(context, deployment_id):
     deployment = software_deployment_get(context, deployment_id)
     session = context.session
-    with session.begin(subtransactions=True):
+    with session.begin():
         session.delete(deployment)
 
 
@@ -1529,8 +1592,7 @@ def sync_point_delete_all_by_stack_and_traversal(context, stack_id,
     return rows_deleted
 
 
-@oslo_db_api.wrap_db_retry(max_retries=3, retry_on_deadlock=True,
-                           retry_interval=0.5, inc_retry_interval=True)
+@retry_on_db_error
 def sync_point_create(context, values):
     values['entity_id'] = str(values['entity_id'])
     sync_point_ref = models.SyncPoint()
@@ -1546,6 +1608,7 @@ def sync_point_get(context, entity_id, traversal_id, is_update):
     )
 
 
+@retry_on_db_error
 def sync_point_update_input_data(context, entity_id,
                                  traversal_id, is_update, atomic_key,
                                  input_data):
@@ -1589,7 +1652,7 @@ def _db_encrypt_or_decrypt_template_params(
         batch_size=batch_size)
     next_batch = list(itertools.islice(template_batches, batch_size))
     while next_batch:
-        with session.begin(subtransactions=True):
+        with session.begin():
             for raw_template in next_batch:
                 try:
                     if verbose:
@@ -1620,7 +1683,7 @@ def _db_encrypt_or_decrypt_template_params(
                                     not param_schemata[param_name].hidden):
                                 continue
                             encrypted_val = crypt.encrypt(
-                                six.text_type(param_val), encryption_key)
+                                str(param_val), encryption_key)
                             env['parameters'][param_name] = encrypted_val
                             encrypted_params.append(param_name)
                             needs_update = True
@@ -1672,7 +1735,7 @@ def _db_encrypt_or_decrypt_resource_prop_data_legacy(
         batch_size=batch_size)
     next_batch = list(itertools.islice(resource_batches, batch_size))
     while next_batch:
-        with session.begin(subtransactions=True):
+        with session.begin():
             for resource in next_batch:
                 if not resource.properties_data:
                     continue
@@ -1719,7 +1782,7 @@ def _db_encrypt_or_decrypt_resource_prop_data(
         model=models.ResourcePropertiesData, batch_size=batch_size)
     next_batch = list(itertools.islice(rpd_batches, batch_size))
     while next_batch:
-        with session.begin(subtransactions=True):
+        with session.begin():
             for rpd in next_batch:
                 if not rpd.data:
                     continue
@@ -1759,7 +1822,7 @@ def db_encrypt_parameters_and_properties(ctxt, encryption_key, batch_size=50,
     :param ctxt: RPC context
     :param encryption_key: key that will be used for parameter and property
                            encryption
-    :param batch_size: number of templates requested from db in each iteration.
+    :param batch_size: number of templates requested from DB in each iteration.
                        50 means that heat requests 50 templates, encrypt them
                        and proceed with next 50 items.
     :param verbose: log an INFO message when processing of each raw_template or
@@ -1783,7 +1846,7 @@ def db_decrypt_parameters_and_properties(ctxt, encryption_key, batch_size=50,
     :param ctxt: RPC context
     :param encryption_key: key that will be used for parameter and property
                            decryption
-    :param batch_size: number of templates requested from db in each iteration.
+    :param batch_size: number of templates requested from DB in each iteration.
                        50 means that heat requests 50 templates, encrypt them
                        and proceed with next 50 items.
     :param verbose: log an INFO message when processing of each raw_template or
@@ -1804,7 +1867,7 @@ def db_properties_data_migrate(ctxt, batch_size=50):
     """Migrate properties data from legacy columns to new location in db.
 
     :param ctxt: RPC context
-    :param batch_size: number of templates requested from db in each iteration.
+    :param batch_size: number of templates requested from DB in each iteration.
                        50 means that heat requests 50 templates, encrypt them
                        and proceed with next 50 items.
     """
